﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;

using MySql.Data.MySqlClient;

namespace MySqlAutoIncrBug {
    
    class Program {

        private const int TaskCount = 2;
        private const string MySqlService = "MySQL56-1";
        private const string ConnectionString = "Server=localhost;Database=ta;User ID=root;Password=supersecret;";

        private const string DropCreateDdl =

@"DROP TABLE IF EXISTS `test`;
CREATE TABLE `test` (
  `test_id` int(11) NOT NULL AUTO_INCREMENT,
  `testcol` varchar(256) COLLATE utf8_bin NOT NULL,
  PRIMARY KEY (`test_id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_bin;";

        static void Main(string[] args) {

            SetupDb();
            RestartMySql();

            Barrier barrier = new Barrier(TaskCount);
            IEnumerable<Task> tasks = Enumerable.Range(0, TaskCount)
                .Select(_ => Task.Run(() => InsertOne(barrier)));
            Task.WaitAll(tasks.ToArray());

            Console.WriteLine("ENTER to exit");
            Console.ReadLine();
        }

        private static void InsertOne(Barrier barrier) {

            int threadId = Thread.CurrentThread.ManagedThreadId;
            Console.WriteLine("[{0}] Executing", threadId);
            try {
                using (MySqlConnection con = GetOpenCon()) {
                    using (MySqlCommand cmd = GetInsertCmd(con)) {
                        barrier.SignalAndWait();
                        cmd.ExecuteNonQuery();
                        Console.WriteLine("[{0}] Insert success, id: {1}", threadId, cmd.LastInsertedId);
                    }
                }
            } catch (Exception e) {
                Console.WriteLine("[{0}] Exception: {1}", threadId, e.Message);
            }
        }

        private static MySqlConnection GetOpenCon() {

            MySqlConnection answer = new MySqlConnection(ConnectionString);
            answer.Open();
            return answer;
        }

        private static MySqlCommand GetInsertCmd(MySqlConnection con) {

            return new MySqlCommand("INSERT INTO `test` (`testcol`) VALUES ('aldsldjfhasjk');", con);
        }

        private static void SetupDb() {

            ulong autoIncLockMode, autoIncInc;

            using (MySqlConnection con = GetOpenCon()) {
                using (MySqlCommand lockModeCmd = new MySqlCommand("SELECT @@GLOBAL.innodb_autoinc_lock_mode;", con)) autoIncLockMode = (ulong)lockModeCmd.ExecuteScalar();
                using (MySqlCommand incIncCmd = new MySqlCommand("SELECT @@GLOBAL.auto_increment_increment;", con)) autoIncInc = (ulong)incIncCmd.ExecuteScalar();
                using (MySqlCommand ddlCmd = new MySqlCommand(DropCreateDdl, con)) ddlCmd.ExecuteNonQuery();
                using (MySqlCommand insertCmd = GetInsertCmd(con)) insertCmd.ExecuteNonQuery();
            }

            Console.WriteLine("This application should be able to reproduce the bug http://bugs.mysql.com/bug.php?id=76872");
            Console.WriteLine("provided innodb_autoinc_lock_mode > 0 and auto_increment_increment > 1. Current values: {0}, {1}", autoIncLockMode, autoIncInc);
            if (autoIncLockMode == 0 || autoIncInc < 2) {
                Console.WriteLine("Set the appropriate values in my.ini to reproduce the problem");
            }
            Console.WriteLine("The problem has been reproduced if at least one thread report an exception with a 'Duplicate entry ...' message");
            Console.WriteLine();
        }

        private static void RestartMySql() {

            foreach (string cmd in new string[] { "stop", "start" }) {
                Process proc = new Process {
                    StartInfo = new ProcessStartInfo {
                        FileName = "net",
                        Arguments = cmd + " " + MySqlService,
                        UseShellExecute = false,
                        CreateNoWindow = true,
                    }
                };
                proc.Start();
                proc.WaitForExit();
            }
        }
    }
}
